"""Function for the Levenshtein algorithm.

Note: This Python implementation is very inefficient. Please use this C++
implementation instead: https://github.com/wq2012/word_levenshtein
"""
import numpy as np
from enum import Enum


class EditOp(Enum):
  Correct = 0
  Substitution = 1
  Insertion = 2
  Deletion = 3


# Computes the Levenshtein alignment between strings ref and hyp, where the
# tokens in each string are separated by delimiter.
# Outputs a tuple : (edit_distance, alignment) where
# alignment is a list of pairs (ref_pos, hyp_pos) where ref_pos is a position
# in ref and hyp_pos is a position in hyp.
# As an example, for strings 'a b' and 'a c', the output would look like:
# (1, [(0,0), (1,1)]
# Note that insertions are represented as (-1, j) and deletions as (i, -1).
def levenshtein_with_edits(
    ref: str,
    hyp: str,
    print_debug_info: bool = False) -> tuple[int, list[tuple[int, int]]]:
  align = []
  s1 = ref.split()
  s2 = hyp.split()
  n1 = len(s1)
  n2 = len(s2)
  costs = np.zeros((n1+1, n2+1), dtype=np.int32)
  backptr = np.zeros((n1+1, n2+1), dtype=EditOp)

  for i in range(n1+1):  # ref
    costs[i][0] = i  # deletions

  for j in range(n2):  # hyp
    costs[0][j+1] = j+1  # insertions
    for i in range(n1):  # ref
      # (i,j) <- (i,j-1)
      ins = costs[i+1][j] + 1
      # (i,j) <- (i-1,j)
      del_ = costs[i][j+1] + 1
      # (i,j) <- (i-1,j-1)
      sub = costs[i][j] + (s1[i] != s2[j])
      costs[i + 1][j + 1] = min(ins, del_, sub)
      if (costs[i+1][j+1] == ins):
        backptr[i+1][j+1] = EditOp.Insertion
      elif (costs[i+1][j+1] == del_):
        backptr[i+1][j+1] = EditOp.Deletion
      elif (s1[i] == s2[j]):
        backptr[i+1][j+1] = EditOp.Correct
      else:
        backptr[i+1][j+1] = EditOp.Substitution

  if print_debug_info:
    print("Mincost: ", costs[n1][n2])
  i = n1
  j = n2
  # Emits pairs (n1_pos, n2_pos) where n1_pos is a position in n1 and n2_pos
  # is a position in n2.
  while (i > 0 or j > 0):
    if print_debug_info:
      print("i: ", i, " j: ", j)
    ed_op = EditOp.Correct
    if (i >= 0 and j >= 0):
      ed_op = backptr[i][j]
    if (i >= 0 and j < 0):
      ed_op = EditOp.Deletion
    if (i < 0 and j >= 0):
      ed_op = EditOp.Insertion
    if (i < 0 and j < 0):
      raise RuntimeError("Invalid alignment")
    if (ed_op == EditOp.Insertion):
      align.append((-1, j-1))
      j -= 1
    elif (ed_op == EditOp.Deletion):
      align.append((i-1, -1))
      i -= 1
    else:
      align.append((i-1, j-1))
      i -= 1
      j -= 1

  align.reverse()
  return costs[n1][n2], align



"""A Python implementation of the ASR and diarization metrics.

We implement:
- Word Error Rate (WER):  This is the most widely used metrics for
  Automatic Speech Recognition (ASR). See
  https://en.wikipedia.org/wiki/Word_error_rate
- Word Diarization Error Rate (WDER): This metric was proposed by Google. See
  Shafey, Laurent El, Hagen Soltau, and Izhak Shafran. "Joint speech
  recognition and speaker diarization via sequence transduction."
  arXiv preprint arXiv:1907.05337 (2019).
  https://arxiv.org/pdf/1907.05337
- Concatenated minimum-permutation Word Error Rate (cpWER): This metric was
  used in the CHiME-6 Challenge. See
  Watanabe, Shinji, et al. "CHiME-6 challenge: Tackling multispeaker speech
  recognition for unsegmented recordings."
  arXiv preprint arXiv:2004.09249 (2020).
  https://arxiv.org/pdf/2004.09249

Note: This implementation is different from Google's internal implementation
that we used in the paper, but is a best-effort attempt to replicate the
results. The biggest differences are from text normalization, such as
de-punctuation.
"""

import dataclasses
from typing import Any, Optional
import numpy as np
from scipy import optimize
import tqdm
from diarizationlm import utils
import word_levenshtein as levenshtein


@dataclasses.dataclass
class UtteranceMetrics:
  """Metrics for one utterance."""

  wer_insert: int = 0
  wer_delete: int = 0
  wer_sub: int = 0
  wer_correct: int = 0
  wer_total: int = 0

  wder_sub: int = 0
  wder_correct: int = 0
  wder_total: int = 0

  cpwer_insert: int = 0
  cpwer_delete: int = 0
  cpwer_sub: int = 0
  cpwer_correct: int = 0
  cpwer_total: int = 0


def merge_cpwer(
    wer_metrics: list[UtteranceMetrics], cpwer_metrics: UtteranceMetrics
) -> None:
  """Compute cpWER metrics by merging a list of WER metrics."""
  for utt in wer_metrics:
    cpwer_metrics.cpwer_insert += utt.wer_insert
    cpwer_metrics.cpwer_delete += utt.wer_delete
    cpwer_metrics.cpwer_sub += utt.wer_sub
    cpwer_metrics.cpwer_correct += utt.wer_correct
    cpwer_metrics.cpwer_total += utt.wer_total


def compute_wer(
    hyp_text: str, ref_text: str
) -> tuple[UtteranceMetrics, list[tuple[int, int]]]:
  """Compute the word error rate of an utterance."""
  result = UtteranceMetrics()
  hyp_normalized = utils.normalize_text(hyp_text)
  ref_normalized = utils.normalize_text(ref_text)
  hyp_words = hyp_normalized.split()
  ref_words = ref_normalized.split()

  # Get the alignment.
  _, align = levenshtein.levenshtein_with_edits(ref_normalized, hyp_normalized)

  # Apply the alignment on ref speakers.
  for i, j in align:
    if i == -1:
      result.wer_insert += 1
    elif j == -1:
      result.wer_delete += 1
    else:
      if ref_words[i] == hyp_words[j]:
        result.wer_correct += 1
      else:
        result.wer_sub += 1

  result.wer_total = result.wer_correct + result.wer_sub + result.wer_delete
  assert result.wer_total == len(ref_words)
  return result, align


def compute_utterance_metrics(
    hyp_text: str,
    ref_text: str,
    hyp_spk: Optional[str] = None,
    ref_spk: Optional[str] = None,
) -> UtteranceMetrics:
  """Compute all metrics of an utterance."""
  hyp_normalized = utils.normalize_text(hyp_text)
  ref_normalized = utils.normalize_text(ref_text)
  hyp_words = hyp_normalized.split()
  ref_words = ref_normalized.split()

  ########################################
  # Compute WER.
  ########################################
  result, align = compute_wer(hyp_text, ref_text)

  compute_diarization_metrics = hyp_spk or ref_spk
  if not compute_diarization_metrics:
    return result

  if not (hyp_spk and ref_spk):
    raise ValueError("hyp_spk and ref_spk must be both unset or both set.")

  ########################################
  # Compute WDER.
  ########################################
  hyp_spk_list = [int(x) for x in hyp_spk.split()]
  ref_spk_list = [int(x) for x in ref_spk.split()]
  if len(hyp_spk_list) != len(hyp_words):
    raise ValueError("hyp_spk and hyp_text must have the same length.")
  if len(ref_spk_list) != len(ref_words):
    raise ValueError("ref_spk and ref_text must have the same length.")
  hyp_spk_list_align = []
  ref_spk_list_align = []

  for i, j in align:
    if i != -1 and j != -1:
      ref_spk_list_align.append(ref_spk_list[i])
      hyp_spk_list_align.append(hyp_spk_list[j])

  # Build cost matrix.
  max_spk = max(max(ref_spk_list_align), max(hyp_spk_list_align))
  cost_matrix = np.zeros((max_spk, max_spk), dtype=int)
  for aligned, original in zip(ref_spk_list_align, hyp_spk_list_align):
    cost_matrix[aligned - 1, original - 1] += 1

  # Solve alignment.
  row_index, col_index = optimize.linear_sum_assignment(
      cost_matrix, maximize=True
  )
  result.wder_correct = int(cost_matrix[row_index, col_index].sum())
  result.wder_total = len(ref_spk_list_align)
  result.wder_sub = result.wder_total - result.wder_correct

  ########################################
  # Compute cpWER.
  ########################################
  spk_pair_metrics = {}
  cost_matrix = np.zeros((max_spk, max_spk), dtype=int)
  for i in range(1, max_spk + 1):
    ref_words_for_spk = [
        ref_words[k] for k in range(len(ref_words)) if ref_spk_list[k] == i
    ]
    if not ref_words_for_spk:
      continue
    for j in range(1, max_spk + 1):
      hyp_words_for_spk = [
          hyp_words[k] for k in range(len(hyp_words)) if hyp_spk_list[k] == j
      ]
      if not hyp_words_for_spk:
        continue
      spk_pair_metrics[(i, j)], _ = compute_wer(
          hyp_text=" ".join(hyp_words_for_spk),
          ref_text=" ".join(ref_words_for_spk),
      )
      cost_matrix[i - 1, j - 1] = spk_pair_metrics[(i, j)].wer_correct

  # Solve alignment.
  row_index, col_index = optimize.linear_sum_assignment(
      cost_matrix, maximize=True
  )
  metrics_to_concat = []
  for r, c in zip(row_index, col_index):
    if (r + 1, c + 1) not in spk_pair_metrics:
      continue
    metrics_to_concat.append(spk_pair_metrics[(r + 1, c + 1)])
  merge_cpwer(metrics_to_concat, result)
  return result


def compute_metrics_on_json_dict(
    json_dict: dict[str, Any],
    ref_text_field: str = "ref_text",
    hyp_text_field: str = "hyp_text",
    ref_spk_field: str = "ref_spk",
    hyp_spk_field: str = "hyp_spk",
) -> dict[str, Any]:
  """Compute metrics for all utterances in a json object."""
  compute_diarization_metrics = ref_spk_field or hyp_spk_field
  if compute_diarization_metrics:
    if not (ref_spk_field and hyp_spk_field):
      raise ValueError(
          "hyp_spk_field and ref_spk_field must be both unset or both set."
      )
  result_dict = {
      "utterances": [],
  }
  for utt in tqdm.tqdm(json_dict["utterances"]):
    if compute_diarization_metrics:
      utt_metrics = compute_utterance_metrics(
          hyp_text=utt[hyp_text_field],
          ref_text=utt[ref_text_field],
          hyp_spk=utt[hyp_spk_field],
          ref_spk=utt[ref_spk_field],
      )
    else:
      utt_metrics = compute_utterance_metrics(
          hyp_text=utt[hyp_text_field],
          ref_text=utt[ref_text_field],
      )
    utt_result = dataclasses.asdict(utt_metrics)
    utt_result["utterance_id"] = utt["utterance_id"]
    result_dict["utterances"].append(utt_result)

  final_wer_total = 0
  final_wer_correct = 0
  final_wer_sub = 0
  final_wer_delete = 0
  final_wer_insert = 0
  final_wder_total = 0
  final_wder_correct = 0
  final_wder_sub = 0
  final_cpwer_total = 0
  final_cpwer_correct = 0
  final_cpwer_sub = 0
  final_cpwer_delete = 0
  final_cpwer_insert = 0
  for utt in result_dict["utterances"]:
    final_wer_total += utt["wer_total"]
    final_wer_correct += utt["wer_correct"]
    final_wer_sub += utt["wer_sub"]
    final_wer_delete += utt["wer_delete"]
    final_wer_insert += utt["wer_insert"]
    if compute_diarization_metrics:
      final_wder_total += utt["wder_total"]
      final_wder_correct += utt["wder_correct"]
      final_wder_sub += utt["wder_sub"]
      final_cpwer_total += utt["cpwer_total"]
      final_cpwer_correct += utt["cpwer_correct"]
      final_cpwer_sub += utt["cpwer_sub"]
      final_cpwer_delete += utt["cpwer_delete"]
      final_cpwer_insert += utt["cpwer_insert"]

  result_dict["WER"] = (
      final_wer_sub + final_wer_delete + final_wer_insert
  ) / final_wer_total

  if compute_diarization_metrics:
    result_dict["WDER"] = final_wder_sub / final_wder_total
    result_dict["cpWER"] = (
        final_cpwer_sub + final_cpwer_delete + final_cpwer_insert
    ) / final_cpwer_total
  return result_dict



"""Tests for metrics."""

import json
import os
from diarizationlm import metrics
import unittest


class MetricsTest(unittest.TestCase):

  def test_wer_same_utt(self):
    hyp = "hello good morning how are you"
    ref = "Hello. Good morning, how are you?"
    result = metrics.compute_utterance_metrics(hyp, ref)
    self.assertEqual(result.wer_insert, 0)
    self.assertEqual(result.wer_delete, 0)
    self.assertEqual(result.wer_sub, 0)
    self.assertEqual(result.wer_correct, 6)
    self.assertEqual(result.wer_total, 6)

  def test_wer_diff_utt(self):
    hyp = "hi morning how are you doing sir"
    ref = "Hello. Good morning, how are you"
    result = metrics.compute_utterance_metrics(hyp, ref)
    self.assertEqual(result.wer_insert, 2)
    self.assertEqual(result.wer_delete, 1)
    self.assertEqual(result.wer_sub, 1)
    self.assertEqual(result.wer_correct, 4)
    self.assertEqual(result.wer_total, 6)

  def test_wer_special(self):
    ref = "a b c"
    hyp = "x y a c"
    result = metrics.compute_utterance_metrics(hyp, ref)
    self.assertEqual(result.wer_insert, 1)
    self.assertEqual(result.wer_delete, 0)
    self.assertEqual(result.wer_sub, 2)
    self.assertEqual(result.wer_correct, 1)
    self.assertEqual(result.wer_total, 3)

  def test_wder_same_words(self):
    hyp = "hello good morning how are you"
    ref = "Hello. Good morning, how are you?"
    hyp_spk = "1 1 1 2 2 2"
    ref_spk = "2 2 2 2 1 1"
    result = metrics.compute_utterance_metrics(hyp, ref, hyp_spk, ref_spk)
    self.assertEqual(result.wer_correct, 6)
    self.assertEqual(result.wder_sub, 1)
    self.assertEqual(result.wder_correct, 5)
    self.assertEqual(result.wder_total, 6)
    self.assertEqual(result.cpwer_insert, 1)
    self.assertEqual(result.cpwer_delete, 1)
    self.assertEqual(result.cpwer_sub, 0)
    self.assertEqual(result.cpwer_correct, 5)
    self.assertEqual(result.cpwer_total, 6)

  def test_wder_diff_words(self):
    hyp = "a b c d e f g h"
    ref = "a bb c e f gg g h ii"
    hyp_spk = "1 1 1 2 2 2 3 2"
    ref_spk = "2 2 2 2 3 3 4 3 2"
    result = metrics.compute_utterance_metrics(hyp, ref, hyp_spk, ref_spk)
    self.assertEqual(result.wer_correct, 6)
    self.assertEqual(result.wder_sub, 1)
    self.assertEqual(result.wder_correct, 6)
    self.assertEqual(result.wder_total, 7)
    self.assertEqual(result.cpwer_insert, 1)
    self.assertEqual(result.cpwer_delete, 2)
    self.assertEqual(result.cpwer_sub, 3)
    self.assertEqual(result.cpwer_correct, 4)
    self.assertEqual(result.cpwer_total, 9)

  def test_cpwer_edge_word_permutation(self):
    hyp = "y x"
    ref = "x y"
    hyp_spk = "2 1"
    ref_spk = "1 2"
    result = metrics.compute_utterance_metrics(hyp, ref, hyp_spk, ref_spk)
    print("result:", result)
    self.assertEqual(result.wer_insert, 1)
    self.assertEqual(result.wer_delete, 1)
    self.assertEqual(result.wer_sub, 0)
    self.assertEqual(result.wer_correct, 1)
    self.assertEqual(result.wer_total, 2)
    self.assertEqual(result.cpwer_insert, 0)
    self.assertEqual(result.cpwer_delete, 0)
    self.assertEqual(result.cpwer_sub, 0)
    self.assertEqual(result.cpwer_correct, 2)
    self.assertEqual(result.cpwer_total, 2)

  def test_compute_metrics_on_json_dict(self):
    json_dict = {
        "utterances": [
            {
                "utterance_id": "utt1",
                "hyp_text": "hello good morning how are you",
                "hyp_spk": "1 1 1 2 2 2",
                "ref_text": "Hello. Good morning, how are you?",
                "ref_spk": "2 2 2 2 1 1",
            },
            {
                "utterance_id": "utt2",
                "hyp_text": "a b c d e f g h",
                "hyp_spk": "1 1 1 2 2 2 3 2",
                "ref_text": "a bb c e f gg g h ii",
                "ref_spk": "2 2 2 2 3 3 4 3 2",
            },
        ]
    }
    result = metrics.compute_metrics_on_json_dict(json_dict)
    self.assertEqual(result["utterances"][0]["utterance_id"], "utt1")
    self.assertEqual(result["utterances"][1]["utterance_id"], "utt2")
    self.assertAlmostEqual(result["WER"], 0.2666, delta=0.001)
    self.assertAlmostEqual(result["WDER"], 0.1538, delta=0.001)
    self.assertAlmostEqual(result["cpWER"], 0.5333, delta=0.001)

  def test_compute_metrics_on_json_dict_wer_only(self):
    json_dict = {
        "utterances": [
            {
                "utterance_id": "utt1",
                "hyp_text": "hello good morning how are you",
                "hyp_spk": "1 1 1 2 2 2",
                "ref_text": "Hello. Good morning, how are you?",
                "ref_spk": "2 2 2 2 1 1",
            },
            {
                "utterance_id": "utt2",
                "hyp_text": "a b c d e f g h",
                "hyp_spk": "1 1 1 2 2 2 3 2",
                "ref_text": "a bb c e f gg g h ii",
                "ref_spk": "2 2 2 2 3 3 4 3 2",
            },
        ]
    }
    result = metrics.compute_metrics_on_json_dict(
        json_dict, ref_spk_field="", hyp_spk_field=""
    )
    self.assertEqual(result["utterances"][0]["utterance_id"], "utt1")
    self.assertEqual(result["utterances"][1]["utterance_id"], "utt2")
    self.assertAlmostEqual(result["WER"], 0.2666, delta=0.001)

  # This is an example where cpWER < WER.
  def test_compute_metrics_on_json_dict_cpwer_word_permutation(self):
    json_dict = {
        "utterances": [
            {
                "utterance_id": "utt1",
                "hyp_text": "y x",
                "hyp_spk": "2 1",
                "ref_text": "x y",
                "ref_spk": "1 2",
            },
        ]
    }
    result = metrics.compute_metrics_on_json_dict(json_dict)
    self.assertEqual(result["utterances"][0]["utterance_id"], "utt1")
    self.assertAlmostEqual(result["WER"], 1.0, delta=0.001)
    self.assertAlmostEqual(result["cpWER"], 0, delta=0.001)

  def test_compute_metrics_on_json_file(self):
    json_file = os.path.join("testdata", "example_data.json")
    with open(json_file, "rt") as f:
      json_dict = json.load(f)
    result = metrics.compute_metrics_on_json_dict(json_dict)
    self.assertEqual(len(result["utterances"]), 2)
    self.assertEqual(result["utterances"][0]["utterance_id"], "en_0638")
    self.assertEqual(result["utterances"][1]["utterance_id"], "en_4157")
    self.assertAlmostEqual(result["WER"], 0.2363, delta=0.001)
    self.assertAlmostEqual(result["WDER"], 0.0437, delta=0.001)
    self.assertAlmostEqual(result["cpWER"], 0.2793, delta=0.001)

  def test_compute_metrics_on_json_file_oracle(self):
    json_file = os.path.join("testdata", "example_data.json")
    with open(json_file, "rt") as f:
      json_dict = json.load(f)
    result = metrics.compute_metrics_on_json_dict(
        json_dict, hyp_spk_field="hyp_spk_oracle"
    )
    self.assertEqual(len(result["utterances"]), 2)
    self.assertEqual(result["utterances"][0]["utterance_id"], "en_0638")
    self.assertEqual(result["utterances"][1]["utterance_id"], "en_4157")
    self.assertAlmostEqual(result["WER"], 0.2363, delta=0.001)
    self.assertAlmostEqual(result["WDER"], 0.0, delta=0.001)
    self.assertAlmostEqual(result["cpWER"], 0.2363, delta=0.001)

  def test_compute_metrics_on_json_file_degraded(self):
    json_file = os.path.join("testdata", "example_data.json")
    with open(json_file, "rt") as f:
      json_dict = json.load(f)
    result = metrics.compute_metrics_on_json_dict(
        json_dict, ref_spk_field="ref_spk_degraded"
    )
    self.assertEqual(len(result["utterances"]), 2)
    self.assertEqual(result["utterances"][0]["utterance_id"], "en_0638")
    self.assertEqual(result["utterances"][1]["utterance_id"], "en_4157")
    self.assertAlmostEqual(result["WER"], 0.2363, delta=0.001)
    self.assertAlmostEqual(result["WDER"], 0.0, delta=0.001)
    self.assertAlmostEqual(result["cpWER"], 0.2363, delta=0.001)


if __name__ == "__main__":
  unittest.main()



"""Test utils."""
import os
import unittest
from diarizationlm import utils
import datasets


class UtilsTest(unittest.TestCase):

  def test_normalize_text(self):
    text = "Hello,  HI, \"how_are_you?\" Good."
    expected = "hello hi howareyou good"
    self.assertEqual(expected, utils.normalize_text(text))

  def test_speakers_transform(self):
    speakers = ["a", "b", "a", "d"]
    expected = ["1", "2", "1", "3"]
    self.assertListEqual(expected, utils.speakers_transform(speakers))

  def test_get_oracle_speakers(self):
    hyp_spk = "1 1 1 1 2 2 2 2"
    hyp_spk_align = "2 2 2 2 1 1 1 1"
    hyp_spk_oracle = utils.get_oracle_speakers(hyp_spk, hyp_spk_align)
    expected = [1, 1, 1, 1, 2, 2, 2, 2]
    self.assertEqual(expected, hyp_spk_oracle)

  def test_transcript_preserving_speaker_transfer(self):
    src_text = "hello good morning hi how are you pretty good"
    src_spk = "1 1 1 2 2 2 2 1 1"
    tgt_text = "hello morning hi hey are you be good"
    tgt_spk = "1 2 2 2 1 1 2 1"
    expected = "1 1 2 2 2 2 1 1"
    transferred_spk = utils.transcript_preserving_speaker_transfer(
        src_text, src_spk, tgt_text, tgt_spk)
    self.assertEqual(expected, transferred_spk)

  def test_ref_to_oracle(self):
    test_data = {
        "hyp_text": "yo hello hi wow great",
        "hyp_spk": "1 2 3 2 1",
        "ref_text": "hello hi hmm wow great",
        "ref_spk": "1 2 2 3 3",
    }
    self.assertEqual("1 2 3 1 1", utils.ref_to_oracle(test_data))

  def test_hyp_to_degraded(self):
    test_data = {
        "hyp_text": "yo hello hi wow great",
        "hyp_spk": "1 2 3 2 1",
        "ref_text": "hello hi hmm wow great",
        "ref_spk": "1 2 2 3 3",
    }
    self.assertEqual("1 2 2 1 3", utils.hyp_to_degraded(test_data))

  def test_create_diarized_text(self):
    word_labels = ["hi", "how", "are", "you", "good"]
    speaker_labels = ["1", "2", "2", "2", "1"]
    po = utils.PromptOptions(speaker_prefix="<spk:", speaker_suffix=">")
    result = utils.create_diarized_text(
        word_labels, speaker_labels, use_new_line=False, po=po
    )
    self.assertEqual("<spk:1> hi <spk:2> how are you <spk:1> good", result)

  def test_extract_text_and_spk(self):
    completions = "hi <spk:2> how are you <spk:4> good <spk:3 hello"
    po = utils.PromptOptions(speaker_prefix="<spk:", speaker_suffix=">")
    text, spk = utils.extract_text_and_spk(completions, po=po)
    self.assertEqual("hi how are you good hello", text)
    self.assertEqual("1 2 2 2 4 3", spk)

  def test_extract_text_and_spk_with_bad_speaker(self):
    completions = "hi <spk:2> how are you <spk:4> good <spk:u> hello"
    po = utils.PromptOptions(speaker_prefix="<spk:", speaker_suffix=">")
    text, spk = utils.extract_text_and_spk(completions, po=po)
    self.assertEqual("hi how are you good hello", text)
    self.assertEqual("1 2 2 2 4 4", spk)

  def test_generate_prompts_1(self):
    utt = {
        "utterance_id": "dummy",
        "hyp_text": "hi how are you good hello",
        "hyp_spk": "1 2 2 2 1 3",
    }
    po = utils.PromptOptions(
        emit_input_length=48,
        prompt_prefix="START ",
        prompt_suffix=" --> ",
        speaker_prefix="<spk:",
        speaker_suffix=">",
    )
    prompts = utils.generate_prompts(utt, po=po)
    expected = [
        "START <spk:1> hi <spk:2> how are --> ",
        "START <spk:2> you --> ",
        "START <spk:1> good <spk:3> hello --> ",
    ]
    self.assertListEqual(expected, prompts)

  def test_generate_prompts_2(self):
    utt = {
        "utterance_id": "dummy",
        "hyp_text": (
            "hi morning hi how are you I am good and you hello everyone how"
            " are you"
        ),
        "hyp_spk": "1 1 2 2 2 2 1 1 1 1 1 3 3 3 3 3",
    }
    po = utils.PromptOptions(
        emit_input_length=36,
        prompt_prefix="",
        prompt_suffix=" -> ",
        speaker_prefix="<s:",
        speaker_suffix=">",
    )
    prompts = utils.generate_prompts(utt, po=po)
    expected = [
        "<s:1> hi morning <s:2> hi how -> ",
        "<s:2> are you <s:1> I am -> ",
        "<s:1> good and you <s:3> hello -> ",
        "<s:3> everyone how are you -> ",
    ]
    self.assertListEqual(expected, prompts)

  def test_generate_prompts_3(self):
    utt = {
        "utterance_id": "en_6047",
        "hyp_text": (
            "Right We cut the Fox network We put it on our table Please put on"
            " the table We cut the Fox network 1 hour condensed version Okay"
            " They have a condensed version They cut out commercials Yeah you"
            " better not unplug that thing You'll be a mega trouble I had The"
            " phone is We left the portable phone upstairs last night So the"
            " battery ran out Okay so I have I have the phone from the living"
            " room in the bedroom with a cord stretch across so I can lay on"
            " the bed of course Yeah So that's why Lee almost tripped over the"
            " phone On recorder here So anyway the backers the packers have the"
            " Packers played yet This weekend don't tell me anything more Okay"
            " I might be able to follow that That game sometimes they replay it"
            " Sometimes they replay the whole thing so you get to watch all"
            " three hours Sometimes Fox replays a 1-hour highlight And so weird"
            " because the first time we watch the 1-hour thing it was like all"
            " right you know so you get all set up and everything like that"
            " Also it's like wait a minute second quarter you All right Yeah"
            " what they do is you know all those long boring plays Yeah you"
            " know like when they go but you know those are kind of fun to see"
            " too but I mean it's better than seeing no game at all but I'd"
            " rather see the whole 3 hours Oh really Well I was there today Oh"
            " that's right It's Sunday night Yeah Oh it's at Lambeau then Yeah"
            " it was that landlord I know I know we won because you said who"
            " would we be today or brat Who would be playing Cincinnati Yeah it"
            " was the second half was good Oh so if you are covered for time"
            " then don't bother watching the first half hour just watch your"
            " second half hour So the second half is exciting huh Yeah it was"
            " good What was an exciting It wasn't an exciting game but was it"
            " the packer Viking No not the pack or Viking I think it was the"
            " pack of beer game Yeah it was a pack of beer game L That was"
            " awesome That was I mean it goes right down to the last two"
            " seconds Lori was Laurie Laurie taped at one night she was"
            " exhausted and fell right asleep and so I told her to put the tape"
            " in and I I couldn't stay up either I was too tired It would have"
            " gone until like 2:00 in the morning Yeah so so I just didn't I"
            " just didn't watch it either so the next night Lori was I think"
            " she fell asleep or something and I watched the whole game and I"
            " could hardly keep myself quiet I know the next day Lori said Well"
            " you know how's the game and I just looked down to the floor and I"
            " said it's good you got to watch it and don't ask me anything else"
            " Then she watch it the next night I had a meeting to go try to get"
            " home till about 1:00 in the morning So I I came home and she was"
            " just finishing watching it How did you keep quiet while you were"
            " watching it I said Oh it was so hard I said did I give it away"
            " when I when I you know when I told you and she said no but you"
            " made me want to watch it I said are you sure I didn't give it"
            " away She said No I can't believe that they goes back to the last"
            " two seconds you know cuz at first you know you figure there's 15"
            " seconds they got I'll go all the way down the field you know this"
            " isn't We close and then they start making it and you go Oh my"
            " gosh I got to get down there We I I got tickets from a parent"
            " that game a parent called me up and gave me tickets today Or was"
            " last week's game Okay and how do you go with this guy or does he"
            " give you the ticket No he gave me the tickets I went with a"
            " teacher friend of mine I am so disappoint said that that I mean"
            " I'm not disappointed at all I can't use that word I take it back"
            " It's so bizarre that you would call me because I just wrote you a"
            " one-page letter because I knew Cuz I knew Dad was going to call"
            " Yeah So I just wrote you this whole one-piece letter Now I feel"
            " like I've already told you everything So anyway here are the"
            " things that I told you the letters Your house No your house is"
            " beautiful Yes we got the pictures the day before we got the film"
            " Oh good We got that We got the video Oh so we So we get the"
            " pictures and my my closest Filipino friend happened to be here"
            " His wife is the one who lived with us the first year we're in"
            " Buchanan and I'm going to these pictures Actually I opened up the"
            " post office This is funny I open them out the post office as in"
            " fact we got the pictures the day we got the film So I'm in the"
            " when you get a package at the post office you have to go to this"
            " different room and then they look for the package for you So I"
            " was in the package I was in the post office room and this other"
            " guy was going to get the package and I get the pictures and the"
            " people are so nosy"
        ),
        "hyp_spk": (
            "1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 1 1 1 1 1 2"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1"
            " 1 1 1 1 2 2 2 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 1 2 2"
            " 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1"
            " 1 2 2 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 1 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 1"
            " 1 1 1 1 1 1 1 1 1 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 2 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1"
        ),
    }
    po = utils.PromptOptions(
        emit_input_length=12000,
        prompt_prefix=(
            "Move potentially misplaced words to the right speaker in the"
            " following transcript.\n"
        ),
        prompt_suffix=" --> ",
    )
    prompts = utils.generate_prompts(utt, po=po)
    for prompt in prompts:
      self.assertGreater(len(prompt), 12000 // 3)
      self.assertLess(len(prompt), 12000)
    self.assertEqual(len(prompts), 1)

  def test_generate_prompts_4(self):
    utt = {
        "utterance_id": "fe_03_00112",
        "hyp_text": (
            "yeah i've i don't really feel that informed about it either it"
            " seems one sided to me no i mean i definitely have the opinion"
            " that if anything is done it should be done uh unil- unilaterally"
            " it should be done through the u._n. because yes u._n. is a"
            " recognized international authority and the u._s. technically is"
            " that's the that's the and yeah and why we never honor the power"
            " that they have i don't understand that either yeah well i mean"
            " basically the united states doesn't have to big powerful country"
            " right yes rather than rather than sin- uh singly which makes us"
            " look even worse yeah i mean no- not only does it make us look"
            " worse it makes the process somewhat unaccountable it makes it a a"
            " war an an aggressive act yes yes um that's not being monitored by"
            " by the authority and i guess i don't understand why now after"
            " after everything last year is is this coming i it's like going to"
            " is last year an excuse for this for this do you think so i don't"
            " know it seems to me do you it seems to me that it's pretty clear"
            " it's maybe something that the bush administration wanted to have"
            " on its agenda when it came in but then oh interesting but then"
            " the change of the mood of the country um after the terrorist"
            " attacks last year um made something possible that maybe the rest"
            " country wouldn't have been ready for before the people weren't"
            " ready at that time to send it would've been a different"
            " atmosphere yeah i think so people would've not been how"
            " interesting ready for war or or sort of chomping at the bit to go"
            " kill someone to go have some revenge on someone even if it's not"
            " the attacker right right the mind set would've been different we"
            " wouldn't have been thinking in that direction probably probably"
            " would've been more sort of i don't know it's a scary it's a scary"
            " part of the world that that i don't understand and i wish i did"
            " more um e- you know that i would've i that i would read more and"
            " understand it more i have a son whose really into it and i should"
            " pay attention more to what he says is he sort of is he just a"
            " person who's interested or is he some sort of just very"
            " interested yeah ver- ah he's not a journalist or uh political"
            " scientist you could say no no just a college student who enjoys"
            " watching b._b._c. and all that yeah but i don't have a i don't"
            " really i guess my opinion would be that i don't have enough of a"
            " a background in it and that i um i'm not sure that we're on the"
            " right course i don't know whether that would that sounds right"
            " doesn't it yeah yeah it seems like we're sort of agreeing very"
            " interesting so yea- we're kind of like what they used to call"
            " what the doves instead of the hawks yeah right i'm aging myself"
            " now see that yes exactly dove i'm sorry i said i'm aging myself"
            " see that oh by saying that they used to be calling it doves"
            " instead of no no no those phrases from from the phr- the phrase"
            " doves and hawks oh because it yes associated with vietnam yes"
            " yeah yeah i've heard it in other contexts i've heard it in the"
            " con- i've heard it in the context of present day israel for"
            " instance oh yes and that election is coming up too i wonder how"
            " that's going to work out with all that's going on i don't know"
            " either well i'm not i really not going to stay on too long is"
            " that okay did we discuss enough well i think actually eh i think"
            " we might need to stay on the full ten minutes to make sure we get"
            " paid are you serious uh i mean it's only like i think ten or"
            " maybe twelve minutes that it goes for but have you have you not"
            " done this before no yeah i did a couple i'm not totally sure but"
            " i would think that they wouldn't want you to just give your i"
            " think they would want you to do the whole time since yo- we are"
            " getting paid i hadn't thought of that i should've probably um"
            " maybe had it more so that it would be at home rather than at work"
            " oh did you have oh you just yeah you're worried that someone's"
            " going to walk by and yeah uh yeah it would maybe maybe could we"
            " finish this call up and yeah i don't know but but i guess if they"
            " if you're supposed to i didn't start timing did you did you no i"
            " didn't i i think we've only i i think we do have a little bit"
            " more time to cover yeah yes i bet we do i def- i i i would like"
            " to finish it just so i can get my money okay okay go head uh so"
            " um it seems that we haven't we don't have a whole lot much more"
            " to say about the iraq situation now do we right you were talking"
            " about the press coverage before i know that like you said we get"
            " a pretty one sided picture probably here in the u._s. my"
            " girlfriend in particular read and i don't in the case of germany"
            " she reads the german presses especially right and i know someone"
            " that does that too and go head i want to hear what your slant is"
            " on that oh well just generally that there's the the you know um"
            " schroeder right is his name the the chancellor who did very well"
            " for himself by opposing right um u._s. military action um and"
            " right it helped his popularity in germany so it's clear that um"
            " in that country at least the uh the mood is definitely again"
            " unilateral and aren't they protesting or something too i don't"
            " know but or or bad feelings i guess yeah that that the way th-"
            " there there stance is the opposite right yeah i think in gen- i"
            " think the majority opinion in fact probably in most european"
            " countries uh with a population at least against u._s. unilateral"
            " action i don't know about against u._n. action right that's and i"
            " think that's probably how they the rest ma- maybe the other"
            " european countries feel the same i don't know though isn- isn't"
            " england still with us i believe that england is but i don't know"
            " what the majority opinion of the population is in en- yeah i i i"
            " think i read somewhere that it's also sort of yes much less uh"
            " committed huh be interesting to just be over there and be a"
            " little fly on the wall and see what they think don't you think"
            " uh-huh yeah i mean i i you'd think actually if you dug around on"
            " the internet you could probably find some polling information"
            " from those countries but i'd like to talk to people yeah in"
            " general wouldn't you rather than just yeah cue from the news or"
            " whatever how do they feel you know but i guess being on a college"
            " campus you could do that too because there's so many"
            " international students that have opinions about how things are"
            " that's definitely true i eh i don't know what your experience is"
            " but but yeah i'm on a college campus i right and and that they"
            " would um maybe weigh in differently than what than what america"
            " would think i have one professor in my department who's very very"
            " vocal on the issue and has a big war uh really statement on his"
            " really door and on his web page really so he's someone who's not"
            " afraid to make a big some other people are probably biting there"
            " tongues more if they're here in the u._s. huh i mean it depends"
            " on where you come from too whether you feel comfortable political"
            " opinion yes huh i i but he's he's he's definitely the other way"
            " huh yeah he definitely is ou- outspoken dove yes uh eh you know"
            " it's interesting that um issues now are not so much about young"
            " people having to go off to war to wear a head band with vietnam"
            " you know the protests about um invol- you know um the peace"
            " protests and and and the draft and all of that it's interesting"
            " now that that doesn't hardly ever comes up as an issue as far as"
            " um you who's going to fight is you know so the two first of all"
            " that there's this issue of yeah as you said who's going to fight"
            " it doesn't seem to come up much and also that it doesn't there"
            " does not the there's not yet at any rate in any sort of right"
            " -mong among could be that you know people of the age that that it"
            " would it really impact and really the other ones would have to go"
            " out and do it you know um you don't hear it there not i don't"
            " think i think they feel insulated or apart from it or something"
            " oh that's somebody else some nameless other person that's not"
            " them it doesn't doesn't effect them you know it could also be"
            " that it actually at this point has wide spread support among"
            " young people oh you think i don't know it seems like i mean my"
            " impression is that hm but um wow i know my my daughter has a"
            " boyfriend who's a marine and i i keep asking if he's going to be"
            " going over but so far no he's not so"
        ),
        "hyp_spk": (
            "1 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2"
            " 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2"
            " 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 2 2 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2"
            " 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1 1 2 2 2 1 1 1 1 2 2 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 2 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 2 2 2"
            " 2 2 1 1 1 1 1 1 1 1 2 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 2 2 2"
            " 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2"
            " 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2"
            " 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2"
            " 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 2 2 2"
            " 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 2 2 2"
            " 2 2 2 2 1 1 1 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1"
            " 1 2 1 1 1 2 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2"
            " 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2"
            " 2 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 2 2 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1"
            " 1 1 2 1 1 1 1 1 1 1 1 1 2 1 1 1 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2"
            " 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 1 2 2 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 1 1 1 1 1 1 2 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2"
            " 2 1 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2"
            " 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 1"
            " 1 1 1 1 1 1 1 1 1 1 1 2 1 1 1 1 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2"
            " 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 2 2 2 2 2 2 2 1 2 2 2 2 2 2 2 1 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2"
            " 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1 2 2 2 2 2 2 2 2 2"
            " 2 2 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 1 1 1 2 1 1 1 1 1 1 2 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 2 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 2"
            " 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2"
            " 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 1 2 2 2 2 2 2"
            " 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 1 1 1 1 1 1 1 1 1 2 1 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2"
            " 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2"
            " 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1 1 1 1"
            " 1 1 1 1 2 2 2 1 1 1 1 1 1 1 1 1 1 1 1 2 1 1 2 2 2 2 2 2 2 2 2 2 2"
            " 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2"
        ),
    }
    po = utils.PromptOptions(
        emit_input_length=12000,
        prompt_prefix=(
            "Move potentially misplaced words to the right speaker in the"
            " following transcript.\n"
        ),
        prompt_suffix=" --> ",
    )
    prompts = utils.generate_prompts(utt, po=po)
    for prompt in prompts:
      self.assertGreater(len(prompt), 12000 // 3)
      self.assertLess(len(prompt), 12000)
    self.assertEqual(len(prompts), 1)

  def test_json_utterance_reader(self):
    json_file = os.path.join(
        "testdata/example_data.json"
    )
    # The length limit for both prompt and target.
    max_len = 150
    po = utils.PromptOptions(
        emit_input_length=max_len,
        emit_target_length=max_len,
        prompt_prefix="",
        prompt_suffix=" --> ",
        completion_suffix=" END",
        speaker_prefix="<s:",
        speaker_suffix=">",
    )
    reader = utils.JsonUtteranceReader(
        json_files=json_file,
        text_field="hyp_text",
        input_speaker_field="hyp_spk",
        target_speaker_field="hyp_spk_oracle",
        po=po,
    )
    keys = []
    for key, prompt, target in reader.generate_data_tuple():
      keys.append(key)

      self.assertTrue(prompt.startswith("<s:"))
      self.assertTrue(prompt.endswith(" --> "))
      self.assertTrue(target.startswith("<s:"))
      self.assertTrue(target.endswith(" END"))

      self.assertGreater(len(prompt), max_len / 3)
      self.assertGreater(len(target), max_len / 3)
      self.assertLessEqual(len(prompt), max_len)
      self.assertLessEqual(len(target), max_len)

    self.assertEqual(len(keys), 195)
    self.assertEqual("en_0638_seg0", keys[0])
    self.assertEqual("en_0638_seg1", keys[1])

  def test_dataset_from_generator(self):
    json_file = os.path.join(
        "testdata/example_data.json"
    )
    # The length limit for both prompt and target.
    max_len = 150
    po = utils.PromptOptions(
        emit_input_length=max_len,
        emit_target_length=max_len,
        prompt_prefix="",
        prompt_suffix=" --> ",
        completion_suffix=" END",
        speaker_prefix="<s:",
        speaker_suffix=">",
    )
    reader = utils.JsonUtteranceReader(
        json_files=json_file,
        text_field="hyp_text",
        input_speaker_field="hyp_spk",
        target_speaker_field="hyp_spk_oracle",
        po=po,
    )
    ds = datasets.Dataset.from_generator(reader.generate_data_dict)
    entry = ds[0]
    self.assertEqual(entry["uttid"], "en_0638_seg0")

  def test_find_utt_dict(self):
    data_dict = {
        "utterances": [
            {
                "utterance_id": "utt1",
                "hyp_text": "how are you",
            },
            {
                "utterance_id": "utt2",
                "hyp_text": "good morning",
            },
        ]
    }
    result = utils.find_utt_dict("utt2", data_dict)
    self.assertEqual("good morning", result["hyp_text"])

  def test_update_hyp_text_in_utt_dict(self):
    utt_dict = {
        "utterance_id": "utt1",
        "hyp_text": "how are you good morning",
        "hyp_spk": "1 1 1 2 2",
    }

    new_hyp_spk = "hello how are you good morning hi"
    updated = utils.update_hyp_text_in_utt_dict(utt_dict, new_hyp_spk)
    self.assertEqual("1 1 1 1 2 2 2", updated["hyp_spk"])

  def test_postprocess_completions_for_utt(self):
    utt_dict = {
        "utterance_id": "utt1",
        "hyp_text": "how are you good morning",
        "hyp_spk": "1 1 2 2 2",
        "completions": [
            "<speaker:1> hello how are you",
            "<speaker:2> good morning hi there",
        ],
    }
    utils.postprocess_completions_for_utt(utt_dict)
    self.assertEqual(
        "hello how are you good morning hi there", utt_dict["llm_text"]
    )
    self.assertEqual("1 1 1 1 2 2 2 2", utt_dict["llm_spk"])
    self.assertEqual("1 1 1 2 2", utt_dict["hyp_spk_llm"])

  def test_postprocess_completions_for_utt_with_suffix1(self):
    utt_dict = {
        "utterance_id": "utt1",
        "hyp_text": "how are you good morning",
        "hyp_spk": "1 1 2 2 2",
        "completions": [
            "<speaker:1> hello how are you END",
            "<speaker:2> good morning hi there END",
        ],
    }
    po = utils.PromptOptions(completion_suffix=" END")
    utils.postprocess_completions_for_utt(utt_dict, po=po)
    self.assertEqual(
        "hello how are you good morning hi there", utt_dict["llm_text"]
    )
    self.assertEqual("1 1 1 1 2 2 2 2", utt_dict["llm_spk"])
    self.assertEqual("1 1 1 2 2", utt_dict["hyp_spk_llm"])

  def test_postprocess_completions_for_utt_with_suffix2(self):
    utt_dict = {
        "utterance_id": "utt1",
        "hyp_text": "how are you good morning",
        "hyp_spk": "1 1 2 2 2",
        "completions": [
            "<speaker:1> hello how are you END\t\t\t\t",
            "<speaker:2> good morning hi there END 4 ds 32 df2 fd END sdf ",
        ],
    }
    po = utils.PromptOptions(completion_suffix=" END")
    utils.postprocess_completions_for_utt(utt_dict, po=po)
    self.assertEqual(
        "hello how are you good morning hi there", utt_dict["llm_text"]
    )
    self.assertEqual("1 1 1 1 2 2 2 2", utt_dict["llm_spk"])
    self.assertEqual("1 1 1 2 2", utt_dict["hyp_spk_llm"])

  def test_truncate_suffix_and_tailing_text(self):
    self.assertEqual(
        "hello how are you [eod]",
        utils.truncate_suffix_and_tailing_text(
            text="hello how are you [eod]", suffix=""
        ),
    )
    self.assertEqual(
        "hello how are you",
        utils.truncate_suffix_and_tailing_text(
            text="hello how are you", suffix="[eod]"
        ),
    )
    self.assertEqual(
        "hello how are you",
        utils.truncate_suffix_and_tailing_text(
            text="hello how are you [eod] \t\t\t\t fafsdfdsafsd",
            suffix=" [eod]",
        ),
    )

  def test_transfer_llm_completion(self):
    llm_completion = (
        "19:00 <speaker:1> Hello, how are you doing today? <speaker:2> I am"
        " doing well. What about you? <speaker:1> i'm doing well, too. Thank"
        " you. <speaker:2> my name"
    )
    hyp = (
        "<speaker:1> Hello, how are you doing <speaker:2> today? I am doing"
        " well. What about <speaker:1> you? I'm doing well, too. Thank you."
    )
    transferred = utils.transfer_llm_completion(llm_completion, hyp)
    self.assertEqual(
        "<speaker:1> Hello, how are you doing today? <speaker:2> I am doing"
        " well. What about you? <speaker:1> I'm doing well, too. Thank you.",
        transferred,
    )


if __name__ == "__main__":
  unittest.main()



"""Utility functions and classes."""
from collections.abc import Generator, Sequence
import copy
import dataclasses
import json
import sys
from typing import Any, List, Optional

import numpy as np
from scipy import optimize

import word_levenshtein as levenshtein

PUNCTUATIONS = [",", ".", "_", "?", "!", "-", '"', "'"]


@dataclasses.dataclass
class PromptOptions:
  """Options for generating prompts."""

  # For prompt segmentation.
  emit_input_length: int = 896
  emit_target_length: int = 896

  # Prefix and suffix for prompt and completion.
  # As a reference, OpenAI finetuning API usually suggests:
  # - No prompt prefix
  # - Prompt suffix: " -> "
  # - Completion suffix: " END"
  prompt_prefix: str = ""
  prompt_suffix: str = " --> "
  completion_suffix: str = ""

  # How do we represent the speaker token.
  # We may consider shorter prefix for token efficiency.
  speaker_prefix: str = "<speaker:"
  speaker_suffix: str = ">"


def normalize_text(text: str) -> str:
  """Normalize text."""
  # Convert to lower case.
  text_lower = text.lower().strip()

  # Remove punctuations.
  words = text_lower.split()
  new_words = []
  for word in words:
    new_word = word
    for punc in PUNCTUATIONS:
      replaced = new_word.replace(punc, "")
      if len(replaced.split()) != 1:
        continue
      new_word = replaced
    new_words.append(new_word)
  return " ".join(new_words)


def speakers_transform(speakers: Sequence[str]) -> list[str]:
  """Transform list of speakers to be order based."""
  spk_map = {}
  index = 0
  for spk in speakers:
    if spk not in spk_map:
      index += 1
      spk_map[spk] = index
  return [str(spk_map[spk]) for spk in speakers]


def get_aligned_hyp_speakers(
    hyp_text: str,
    ref_text: str,
    ref_spk: str,
    print_debug_info: bool = False,
) -> str:
  """Align ref_text to hyp_text, then apply the alignment to ref_spk."""
  # Counters for insertions and deletions in hyp and ref text alignment.
  num_insertions, num_deletions = 0, 0

  # Get the alignment.
  _, align = levenshtein.levenshtein_with_edits(
      normalize_text(ref_text), normalize_text(hyp_text)
  )

  ref_spk_list = ref_spk.split()
  hyp_spk_align = []

  # Apply the alignment on ref speakers.
  for i, j in align:
    if i == -1:
      # hyp has insertion
      hyp_spk_align.append("-1")
      num_insertions += 1
    elif j == -1:
      # hyp has deletion
      num_deletions += 1
      continue
    else:
      hyp_spk_align.append(ref_spk_list[i])
  hyp_spk_align = " ".join(hyp_spk_align)

  if print_debug_info:
    print("Number of insertions: ", num_insertions)
    print("Number of deletions: ", num_deletions)
    # This is not the traditional denominator of WER. Instead, this is
    # len(hyp) + len(ref) - len(SUB).
    print("Length of align pairs: ", len(align))
  return hyp_spk_align


def get_oracle_speakers(hyp_spk: str, hyp_spk_align: str) -> Sequence[int]:
  """Get the oracle speakers for hypothesis."""
  hyp_spk_list = [int(x) for x in hyp_spk.split()]
  hyp_spk_align_list = [int(x) for x in hyp_spk_align.split()]

  # Build cost matrix.
  max_spk = max(max(hyp_spk_list), max(hyp_spk_align_list))
  cost_matrix = np.zeros((max_spk, max_spk))
  for aligned, original in zip(hyp_spk_align_list, hyp_spk_list):
    cost_matrix[aligned - 1, original - 1] += 1

  # Solve alignment.
  row_index, col_index = optimize.linear_sum_assignment(
      cost_matrix, maximize=True
  )

  # Build oracle.
  hyp_spk_oracle = hyp_spk_list.copy()
  for i in range(len(hyp_spk_list)):
    if hyp_spk_align_list[i] == -1:
      # There are some missing words. In such cases, we just use the original
      # speaker for these words if possible.
      if hyp_spk_list[i] == -1:
        # If we don't have original speaker for missing words, just use the
        # previous speaker if possible.
        # This is useful for the update_hyp_text_in_utt_dict() function.
        if i == 0:
          hyp_spk_oracle[i] = 1
        else:
          hyp_spk_oracle[i] = hyp_spk_oracle[i - 1]
      continue
    assert row_index[hyp_spk_align_list[i] - 1] == hyp_spk_align_list[i] - 1
    hyp_spk_oracle[i] = col_index[hyp_spk_align_list[i] - 1] + 1

  return hyp_spk_oracle


# Transcript-Preserving Speaker Transfer (TPST)
def transcript_preserving_speaker_transfer(
    src_text: str, src_spk: str, tgt_text: str, tgt_spk: str
) -> str:
  """Apply source speakers to target."""
  if len(tgt_text.split()) != len(tgt_spk.split()):
    raise ValueError("tgt_text and tgt_spk must have the same length")
  if len(src_text.split()) != len(src_spk.split()):
    raise ValueError("src_text and src_spk must have the same length")
  tgt_spk_align = get_aligned_hyp_speakers(
      hyp_text=tgt_text,
      ref_text=src_text,
      ref_spk=src_spk,
  )
  oracle_speakers = get_oracle_speakers(
      hyp_spk=tgt_spk, hyp_spk_align=tgt_spk_align
  )
  return " ".join([str(x) for x in oracle_speakers])


# We can use this to finetune LLM.
# Inputs (prompts): hyp diarized text
# Targets: hyp diarized text with oracle speakers
def ref_to_oracle(json_dict: dict[str, str]) -> str:
  """Apply reference speakers to hypothesis."""
  return transcript_preserving_speaker_transfer(
      src_text=json_dict["ref_text"],
      src_spk=json_dict["ref_spk"],
      tgt_text=json_dict["hyp_text"],
      tgt_spk=json_dict["hyp_spk"],
  )


# Similar to ref_to_oracle, but the opposite direction.
# We can use this to finetune LLM.
# Inputs (prompts): ref diarized text with degraded speakers
# Targets: ref diarized text
def hyp_to_degraded(json_dict: dict[str, str]) -> str:
  """Apply hypothesis speakers to reference."""
  return transcript_preserving_speaker_transfer(
      src_text=json_dict["hyp_text"],
      src_spk=json_dict["hyp_spk"],
      tgt_text=json_dict["ref_text"],
      tgt_spk=json_dict["ref_spk"],
  )


def create_diarized_text(
    word_labels: Sequence[str],
    speaker_labels: Sequence[str],
    use_new_line: bool = False,
    po: PromptOptions = PromptOptions(),
) -> str:
  """Create diarized text from words and speaker labels."""
  output = []
  previous_speaker = None
  for word, speaker in zip(word_labels, speaker_labels):
    if speaker != previous_speaker:
      if previous_speaker and use_new_line:
        output.append("\n")
      output.append(po.speaker_prefix + speaker + po.speaker_suffix)
    output.append(word)
    previous_speaker = speaker
  return " ".join(output)


def extract_text_and_spk(
    completions: str, po: PromptOptions, skip_meaningless_speaker: bool = True
) -> tuple[str, str]:
  """Extract the text and spk from the completions string."""
  spk = "1"
  previous_spk = "1"
  result_text = []
  result_spk = []
  for word in completions.split():
    if word.startswith(po.speaker_prefix):
      if not word.endswith(po.speaker_suffix):
        word += po.speaker_suffix
      spk = word[len(po.speaker_prefix):-len(po.speaker_suffix)]
      # Handle undefined behaviors of non-recognizable spk with a placeholder.
      try:
        spk_int = int(spk)
        if not spk or spk_int < 1 or spk_int > 10:
          raise ValueError("Seeing unexpected word: ", word)
        previous_spk = spk
      except ValueError as exc:
        if skip_meaningless_speaker:
          print("Skipping meaningless speaker token:", word)
          spk = previous_spk
        else:
          raise exc
    else:
      result_text.append(word)
      result_spk.append(spk)
  return " ".join(result_text), " ".join(result_spk)


def discard_empty_str_and_remove_boundary_white_space(
    inputs: List[str],
) -> List[str]:
  return [x.strip() for x in inputs if x.strip()]


@dataclasses.dataclass
class JsonUtteranceReader:
  """Read the json files and generate prompts and targets."""

  json_files: str  # Ignored if utt is given.
  text_field: str
  input_speaker_field: str
  target_speaker_field: str  # If not given, will skip targets.
  po: PromptOptions
  utt: dict[str, str] = dataclasses.field(default_factory=dict)

  def generate_utts(self) -> Generator[dict[str, str], None, None]:
    """Generate an utterance from all json files."""
    if self.utt:
      yield self.utt
      return

    for json_file in self.json_files.split(","):
      with open(json_file) as f:
        data_dict = json.load(f)
        for utt in data_dict["utterances"]:
          yield utt

  def generate_data_tuple(self) -> Generator[tuple[str, str, str], None, None]:
    """Generate uttid-prompt-target tuples."""
    for utt in self.generate_utts():
      yield from self.generate_data_tuple_for_utt(utt)

  def generate_data_dict(self) -> Generator[dict[str, str], None, None]:
    """Generate a dict that can be used for datasets.Dataset.from_generator."""
    for uttid, prompt, target in self.generate_data_tuple():
      yield {"uttid": uttid, "prompt": prompt, "target": target}

  def generate_data_tuple_for_utt(
      self, utt: dict[str, str]
  ) -> Generator[tuple[str, str, str], None, None]:
    """Generate uttid-prompt-target tuples from a single utterance."""
    self.seg_id = 0
    utt_id = utt["utterance_id"]

    # Get the fields from the utterance.
    words = discard_empty_str_and_remove_boundary_white_space(
        utt[self.text_field].split(" ")
    )
    p_speakers = discard_empty_str_and_remove_boundary_white_space(
        utt[self.input_speaker_field].split(" ")
    )
    assert len(words) == len(p_speakers)
    if self.target_speaker_field:
      t_speakers = discard_empty_str_and_remove_boundary_white_space(
          utt[self.target_speaker_field].split(" ")
      )
      assert len(words) == len(t_speakers)
    else:
      t_speakers = []

    yield from self.generate_data_tuple_from_range(
        utt_id, words, p_speakers, t_speakers, start=0, end=len(words)
    )

  def generate_data_tuple_from_range(
      self, utt_id, words, p_speakers, t_speakers, start, end
  ) -> Generator[tuple[str, str, str], None, None]:
    """Generate uttid-prompt-target tuples from a range of words."""
    # Decide whether to call recursively from the estimated length.
    estimated_prompt_length = (
        len(self.po.prompt_prefix)
        + len(" ".join(words[start:end]))
        + len(self.po.prompt_suffix)
    )
    if (
        estimated_prompt_length > self.po.emit_input_length
        or estimated_prompt_length > self.po.emit_target_length
    ):
      yield from self.generate_data_tuple_from_range(
          utt_id, words, p_speakers, t_speakers, start, (start + end) // 2
      )
      yield from self.generate_data_tuple_from_range(
          utt_id, words, p_speakers, t_speakers, (start + end) // 2, end
      )
      return

    prompt = self.po.prompt_prefix
    previous_p_spk = ""
    target = ""
    previous_t_spk = ""

    # Main loop.
    for i in range(start, end):
      word = words[i]
      p_spk = p_speakers[i]
      if p_spk != previous_p_spk:
        if previous_p_spk:
          prompt += " "
        prompt += self.po.speaker_prefix + p_spk + self.po.speaker_suffix
      prompt += " " + word
      previous_p_spk = p_spk

      if self.target_speaker_field:
        t_spk = t_speakers[i]
        if t_spk != previous_t_spk:
          if previous_t_spk:
            target += " "
          target += self.po.speaker_prefix + t_spk + self.po.speaker_suffix
        target += " " + word
        previous_t_spk = t_spk

    prompt_id = utt_id + "_seg" + str(self.seg_id)
    prompt += self.po.prompt_suffix
    target += self.po.completion_suffix
    if (
        len(prompt) <= self.po.emit_input_length
        and len(target) <= self.po.emit_target_length
    ):
      yield (prompt_id, prompt, target)
      self.seg_id += 1
    else:
      yield from self.generate_data_tuple_from_range(
          utt_id, words, p_speakers, t_speakers, start, (start + end) // 2
      )
      yield from self.generate_data_tuple_from_range(
          utt_id, words, p_speakers, t_speakers, (start + end) // 2, end
      )


def generate_prompts(
    utt: dict[str, str],
    po: PromptOptions,
    text_field: str = "hyp_text",
    input_speaker_field: str = "hyp_spk",
) -> list[str]:
  """Generate a list of prompts for a given utt."""
  po_modified = copy.deepcopy(po)
  po_modified.emit_target_length = sys.maxsize
  reader = JsonUtteranceReader(
      json_files="",
      text_field=text_field,
      input_speaker_field=input_speaker_field,
      target_speaker_field="",
      po=po_modified,
      utt=utt,
  )
  prompts = []
  for _, prompt, _ in reader.generate_data_tuple():
    prompts.append(prompt)
  if len(prompts) > 1:
    for prompt in prompts:
      if len(prompt) < po.emit_input_length / 3:
        raise RuntimeError("Prompt too short: ", prompt)
  return prompts


def find_utt_dict(
    utt_id: str, data_dict: dict[str, Any]
) -> Optional[dict[str, str]]:
  """Find a utt_dict with a speicifc utterance_id from data_dict."""
  for utt_dict in data_dict["utterances"]:
    if utt_dict["utterance_id"] == utt_id:
      return utt_dict
  return None


def update_hyp_text_in_utt_dict(
    input_utt_dict: dict[str, str], new_hyp_text
) -> dict[str, str]:
  """Update the hyp_text of a json utt_dict.

  We also transfer its original hyp_spk to the new hyp_text.

  This is useful if we want to use USM ASR transcripts to replace the
  turn-to-diarize transcripts, as the WER of turn-to-diarize transcripts is too
  high.

  Args:
    input_utt_dict: the input utt_dict
    new_hyp_text: the new hyp_text

  Returns:
    the new utt_dict
  """
  utt_dict = copy.deepcopy(input_utt_dict)
  # We don't know the speakers for new_hyp_text, so just use -1 as initial
  # speakers.
  new_hyp_spk = transcript_preserving_speaker_transfer(
      src_text=utt_dict["hyp_text"],
      src_spk=utt_dict["hyp_spk"],
      tgt_text=new_hyp_text,
      tgt_spk=" ".join(["-1" for _ in new_hyp_text.split()]),
  )
  # Update the utt_dict.
  utt_dict["hyp_text"] = new_hyp_text
  utt_dict["hyp_spk"] = new_hyp_spk
  utt_dict["hyp_diarized_text"] = create_diarized_text(
      new_hyp_text.split(), new_hyp_spk.split()
  )
  return utt_dict


def truncate_suffix_and_tailing_text(text: str, suffix: str) -> str:
  """Tailing text after suffix should be removed as well."""
  if suffix and suffix in text:
    return text[: text.find(suffix)]
  return text


def postprocess_completions_for_utt(
    utt: dict[str, Any],
    llm_text_field: str = "llm_text",
    llm_speaker_field: str = "llm_spk",
    transfered_llm_speaker_field: str = "hyp_spk_llm",
    hyp_text_field: str = "hyp_text",
    hyp_spk_field: str = "hyp_spk",
    po: PromptOptions = PromptOptions(),
) -> None:
  """Postprocess the LLM completions of an utterance json dict."""
  # Remove completion suffix if it exists.
  completions_list = []
  for completion in utt["completions"]:
    if po.completion_suffix and po.completion_suffix in completion:
      completion = truncate_suffix_and_tailing_text(
          completion, po.completion_suffix
      )
    completions_list.append(completion)
  completions = " ".join(completions_list).strip()

  # Extract text and speaker.
  utt[llm_text_field], utt[llm_speaker_field] = extract_text_and_spk(
      completions, po=po
  )
  # Tha TPST alignment on LLM output against recognized hypothesis text can be
  # considered as a postprocessing step to ensure the hypothesis text does not
  # change too much from the diarization baseline.
  # Note: this step can arguably be skipped and we directly use LLM output
  # for evaluation. The assumption is LLM does not change original text too
  # much. `update_sstable_with_speakers` should be updated accordingly if so.
  utt[transfered_llm_speaker_field] = transcript_preserving_speaker_transfer(
      src_text=utt[llm_text_field],
      src_spk=utt[llm_speaker_field],
      tgt_text=utt[hyp_text_field],
      tgt_spk=utt[hyp_spk_field],
  )


def transfer_llm_completion(
    llm_completion: str,
    hyp: str,
    po: PromptOptions = PromptOptions(),
) -> str:
  """Transfer the LLM completion text to use text from hypothesis."""
  llm_text, llm_speaker = extract_text_and_spk(
      llm_completion, po=po
  )
  hyp_text, hyp_speaker = extract_text_and_spk(
      hyp, po=po
  )
  transfered_llm_speaker = transcript_preserving_speaker_transfer(
      src_text=llm_text,
      src_spk=llm_speaker,
      tgt_text=hyp_text,
      tgt_spk=hyp_speaker,
  )
  transferred = create_diarized_text(
      word_labels=hyp_text.split(),
      speaker_labels=transfered_llm_speaker.split(),
      po=po,
  )
  return transferred
